#! /usr/bin/env pike77

#charset euc-jp

string romanize(string s)
{
  s = replace(s,
	      ({"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","","",
		"","","","",""}),
	      ({"a","e","i","o","u",
		"sa","se","shi","so","su",
		"za","ze","ji","zo","zu",
		"ka","ke","ki","ko","ku",
		"ga","ge","gi","go","gu",
		"ta","te","chi","to","tsu",
		"da","de","ji","do","zu",
		"ha","he","hi","ho","fu",
		"ba","be","bi","bo","bu",
		"pa","pe","pi","po","pu",
		"ma","me","mi","mo","mu",
		"na","ne","ni","no","nu",
		"ra","re","ri","ro","ru",
		"ya","yo","yu","Ya","Yo",
		"Yu","wa","wo","n","!"})*2);
  s = replace(s,
	      ({"riY", "miY", "hiY", "niY", "chiY", "shiY", "kiY",
		"piY", "biY", "jiY", "giY"}),
	      ({"ry", "my", "hy", "ny", "ch", "sh", "ky",
		"py", "by", "j", "gy"}));
  s = replace(s,
	      ({"!k", "!s", "!t", "!ch", "!p",
		"!.k", "!.s", "!.t", "!.ch", "!.p"}),
	      ({"kk", "ss", "tt", "tch", "pp",
		"k.k", "s.s", "t.t", "t.ch", "p.p"}));
  return s;
}

class Kanji {

  static string kanji;
  static array(string) yomi, english;
  static int jis, unicode, freq, grade, heisig;

  string get_kanji() { return kanji; }
  string get_yomi() { return yomi * " "; }
  string get_english() { return english * " : "; }
  int get_jis() { return jis; }
  int get_grade() { return grade; }
  int get_freq() { return freq; }

  static void create(string s)
  {
    string c="", y="", m="", mm;
    array(string) c2;
    kanji = "";
    english = ({});
    sscanf(s, "%s %x %[A-Za-z0-9 .-]%[^{]%s", kanji, jis, c, y, m);
    sscanf(y, "%sT", y); // Remove nanori and radical readings
    while(sscanf(m, "{%s}%*[ ]%s", mm, m) == 3)
      english += ({ mm });
    if(strlen(kanji) != 1 ||
       sizeof((c2 = c/" "-({""}))) < 1 ||
       sizeof((yomi = y/" "-({""}))) < 1 ||
       sizeof(english) < 1) {
      destruct(this_object());
      return;
    }
    foreach(c2, c)
      switch(c[0]) {
      case 'F':
	freq = (int)c[1..]; break;
      case 'G':
	grade = (int)c[1..]; break;
      case 'U':
	sscanf(c[1..], "%x", unicode); break;
      case 'L':
	heisig = (int)c[1..]; break;	
     }
  }

}

array(Kanji) low_parse_kanjidic(Stdio.Stream s, string fn)
{
  Locale.Charset.Decoder dec = Locale.Charset.decoder("euc-jp");
  string t, buf="";
  array(Kanji) ks = ({});
  int ln = 1;

  while((t = s->read(4096)) && sizeof(t)) {
    buf += dec->feed(t)->drain();
    while(sscanf(buf, "%s\n%s", t, buf) == 2) {
      if(sizeof(t) && t[0] != '#') {
	Kanji k = Kanji(t);
	if(!k)
	  error("Invalid kanji declaration on line %d of %s\n", ln, fn);
	ks += ({ k });
      }
      ln++;
    }
  }

  return ks;
}

array(Kanji) parse_kanjidic(string fn)
{
  return low_parse_kanjidic(Stdio.File(fn, "r"), fn);
}

class KanjiLabel {

  inherit GTK2.Label;

  static Kanji kanji;
  static string mode;
  static int romaji, cheat;

  static void update_text()
  {
    if(!kanji || !mode)
      return;
    string s;
    switch(mode) {
    case "kanji": s = kanji->get_kanji(); break;
    case "yomi": s = kanji->get_yomi(); if(romaji) s = romanize(s); break;
    case "english": s = kanji->get_english(); break;
    }
    set_label(cheat? "> "+s+" <" : s);
  }

  void set_kanji(Kanji k)
  {
    cheat = 0;
    kanji = k;
    update_text();
  }

  void set_mode(string m, int r)
  {
    mode = m;
    romaji = r;
    update_text();
  }

  void set_cheat()
  {
    if(!cheat) {
      cheat = 1;
      update_text();
    }
  }

  Kanji get_kanji()
  {
    return kanji;
  }

};


class KDrillWindow {
  inherit GTK2.Window;

  static array(Kanji) all_kanjis, filtered_kanjis;
  static KanjiLabel klabel;
  static array(KanjiLabel) kbuttons;
  static string show_mode, guess_mode;
  static int randomly, romaji, usefile;
  static GTK2.Statusbar statusbar;
  static mapping(string:int) current_message = ([]);
  static int kanji_index, old_index, answer_index;
  static int max_freq;

  static class OptionsWindow {
    inherit GTK2.Window;

    static int is_shown;

    static void create()
    {
      ::create(GTK2.WindowToplevel);
      set_title("GKDrill options");

      GTK2.RadioButton r;
      GTK2.ToggleButton t;

      GTK2.Vbox v2 = GTK2.Vbox(0,10);

      GTK2.Hbox h = GTK2.Hbox(0,10);

      GTK2.Vbox v = GTK2.Vbox(0,0);
      v->add(r = GTK2.RadioButton("Show kanji")->show());
      r->signal_connect("clicked", lambda() { set_shown("kanji"); });
      if(show_mode == "kanji") r->set_active(1);
      v->add(r = GTK2.RadioButton("Show English", r)->show());
      r->signal_connect("clicked", lambda() { set_shown("english"); });
      if(show_mode == "english") r->set_active(1);
      v->add(r = GTK2.RadioButton("Show meaning", r)->show());
      r->signal_connect("clicked", lambda() { set_shown("yomi"); });
      if(show_mode == "yomi") r->set_active(1);
      h->add(GTK2.Frame()->add(v->show())->show());

      v = GTK2.Vbox(0,10);
      v->add(t = GTK2.ToggleButton("Show randomly")->show());
      t->signal_connect("toggled", lambda(GTK2.ToggleButton w) {
				     set_randomly(w->get_active());
				   });
      if(randomly) t->set_active(1);
      v->add(t = GTK2.ToggleButton("Romaji")->show());
      t->signal_connect("toggled", lambda(GTK2.ToggleButton w) {
				     set_romaji(w->get_active());
				   });
      if(romaji) t->set_active(1);
      v->add(t = GTK2.ToggleButton("Usefile")->show());
      t->signal_connect("toggled", lambda(GTK2.ToggleButton w) {
				     set_usefile(w->get_active());
				   });
      if(usefile) t->set_active(1);
      h->add(v->show());

      v2->add(h->show());

      h = GTK2.Hbox(0,10);

      v = GTK2.Vbox(0,0);
      v->add(r = GTK2.RadioButton("Guess kanji")->show());
      r->signal_connect("clicked", lambda() { set_guess("kanji"); });
      if(guess_mode == "kanji") r->set_active(1);
      v->add(r = GTK2.RadioButton("Guess English", r)->show());
      r->signal_connect("clicked", lambda() { set_guess("english"); });
      if(guess_mode == "english") r->set_active(1);
      v->add(r = GTK2.RadioButton("Guess meaning", r)->show());
      r->signal_connect("clicked", lambda() { set_guess("yomi"); });
      if(guess_mode == "yomi") r->set_active(1);
      h->add(GTK2.Frame()->add(v->show())->show());
 
      v = GTK2.Vbox(0,10);
      v->add(GTK2.Label("Frequency Range")->show());
      GTK2.Hbox h2 = GTK2.Hbox(0,10);
      h2->add(GTK2.SpinButton(1.0,(float)max_freq,1.0)->show());
      h2->add(GTK2.SpinButton(1.0,(float)max_freq,1.0)->show());
      v->add(h2->show());
      h->add(GTK2.Frame()->add(v->show())->show());

      v2->add(h->show());

      add(v2->show());
      signal_connect("delete_event", lambda() { options_window = 0; });
    }
  };

  static OptionsWindow options_window;

  void set_shown(string mode)
  {
    klabel->set_mode(show_mode = mode, romaji);
  }

  void set_guess(string mode)
  {
    kbuttons->set_mode(guess_mode = mode, romaji);
  }

  void set_randomly(int r)
  {
    if(r != randomly) {
      status_message("options", (r? "Will now select kanji randomly" :
				 "Will now go through kanji in order of #"));
      randomly = r;
      if(filtered_kanjis)
	reorder_kanji_list();
    }
  }

  void set_romaji(int r)
  {
    if(r != romaji) {
      status_message("options", (r? "Will now print romaji" :
				 "Will now print kana"));
      romaji = r;
      if(show_mode == "yomi")
	set_shown(show_mode);
      if(guess_mode == "yomi")
	set_guess(guess_mode);
    }
  }

  void set_usefile(int u)
  {
    if(u != usefile) {
      status_message("options", (u? "Now using usefile .kanjiusefile" :
				 "No longer restricting by usefile"));
      usefile = u;
      if(filtered_kanjis)
	apply_kanji_filter();
    }
  }

  static void guess(Kanji k)
  {
    if(k->get_kanji() == klabel->get_kanji()->get_kanji()) {
      status_message("game", "Correct! Now try THIS one...");
      next_kanji();
    } else {
      status_message("game", "Incorrect.");
    }
  }

  static void status_message(string ctx, string msg, mixed ... args)
  {
    if(sizeof(args))
      msg = sprintf(msg, @args);
    int c = statusbar->get_context_id(ctx);
    int id = statusbar->push(c, msg);
    if(!zero_type(current_message[ctx]))
      statusbar->remove(c, current_message[ctx]);
    current_message[ctx] = id;
  }

  static void reorder_kanji_list()
  {
    if(randomly) {
      Array.shuffle(filtered_kanjis);
    } else {
      sort(filtered_kanjis->get_jis(), filtered_kanjis);
    }
  }

  static void next_kanji()
  {
    old_index = kanji_index;
    if(++kanji_index >= sizeof(filtered_kanjis)) {
      kanji_index = 0;
      if(randomly)
	reorder_kanji_list();
    }
    select_kanji();
  }

  static void prev_kanji()
  {
    if(old_index >= 0 && old_index < sizeof(filtered_kanjis)) {
      status_message("game", "Going to previous character");
      int tmp = kanji_index;
      kanji_index = old_index;
      old_index = tmp;
      select_kanji();
    }
  }

  static void select_kanji()
  {
    if(sizeof(filtered_kanjis)<5) {
      status_message("game", "Too few kanji available!");
      return;
    }
    Kanji k = filtered_kanjis[kanji_index];
    klabel->set_kanji(k);
    answer_index = random(5);
    multiset(int) used_indices = (<kanji_index>);
    for(int i=0; i<5; i++)
      if(i == answer_index)
	kbuttons[i]->set_kanji(k);
      else {
	int wrong_index;
	do
	  wrong_index = random(sizeof(filtered_kanjis));
	while(used_indices[wrong_index]);
	kbuttons[i]->set_kanji(filtered_kanjis[wrong_index]);
	used_indices[i] = 1;
      }
  }

  static int filter_kanjis(function(Kanji:int) f)
  {
    filtered_kanjis = filter(all_kanjis, f);
    reorder_kanji_list();
    kanji_index = old_index = -1;
    return sizeof(filtered_kanjis);
  }

  static int apply_kanji_filter()
  {
    return filter_kanjis(lambda(Kanji k) { return k->get_grade() == 2; });
  }

  static void create()
  {
    ::create(GTK2.WindowToplevel);
    set_title("GKDrill");
    all_kanjis = ({});
    all_kanjis += parse_kanjidic("/usr/share/edict/kanjidic");
    max_freq = max(@all_kanjis->get_freq());

    GTK2.Vbox v = GTK2.Vbox(0,0);

    v->add((klabel = KanjiLabel())->show());
    (kbuttons = (({KanjiLabel})*5)())->show();
    foreach(kbuttons, KanjiLabel w) {
      GTK2.Button b = GTK2.Button();
      v->add(b->add(w)->show());
      b->signal_connect("clicked", lambda(GTK2.Button b) {
				     guess(b->get_child()->get_kanji());
				   });
    }

    GTK2.Hbox h = GTK2.Hbox(0,0);
    GTK2.Button b;
    h->add(b = GTK2.Button("Quit")->show());
    b->signal_connect("clicked", lambda() { exit(0); });
    h->add(b = GTK2.Button("Cheat")->show());
    b->signal_connect("clicked", lambda() {
				   status_message("game", "Try Harder!");
				   kbuttons[answer_index]->set_cheat();
				 });
    h->add(b = GTK2.Button("Prev. char")->show());
    b->signal_connect("clicked", prev_kanji);
    h->add(GTK2.Button("Log")->show());
    h->add(b = GTK2.Button("Options...")->show());
    b->signal_connect("clicked", lambda() {
				   if(!options_window)
				     options_window = OptionsWindow();
				   options_window->present();
				 });
    h->add(GTK2.Button("Search...")->show());
    h->add(GTK2.Button("Learn...")->show());
    v->add(h->show());

    v->add(statusbar = GTK2.Statusbar()->show());

    set_shown("english");
    set_guess("kanji");
    set_randomly(1);
    set_romaji(0);
    set_usefile(0);

    apply_kanji_filter();

    foreach(({klabel})+kbuttons, object w)
      w->modify_font(GTK2.PangoFontDescription("normal 20"));

    status_message("main", "Welcome to gkdrill");

    next_kanji();

    add(v->show());
  }
}

int main(int argc, array(string) argv)
{
  random_seed(time());
  GTK2.setup_gtk(argv);
  KDrillWindow()->show()->
    signal_connect("delete_event", lambda() { exit(0); });
  return -1;
}
